"""Converts an OSeMOSYS datafile to a Tabular Data Package

"""
import logging
import os
import sys
from typing import Dict, List

import pandas as pd
from flatten_dict import flatten
from pulp import Amply

from otoole import read_packaged_file
from otoole.preprocess.longify_data import check_datatypes, check_set_datatype, write_out_dataframe

logger = logging.getLogger(__name__)


def convert_file_to_package(path_to_datafile: str, path_to_datapackage: str):
    """Converts an OSeMOSYS datafile to a Tabular Data Package

    Arguments
    ---------
    path_to_datafile: str
    path_to_datapackage: str
        Path to the folder in which to write the new Tabular Data Package

    """
    config = read_packaged_file('config.yaml', 'otoole.preprocess')
    amply_datafile = read_in_datafile(path_to_datafile, config)
    dict_of_dataframes = convert_amply_to_dataframe(amply_datafile, config)
    if not os.path.exists(path_to_datapackage):
        os.mkdir(path_to_datapackage)
    for name, df in dict_of_dataframes.items():
        write_out_dataframe(path_to_datapackage, name, df)
    datapackage = read_packaged_file('datapackage.json', 'otoole.preprocess')
    filepath = os.path.join(path_to_datapackage, 'datapackage.json')
    with open(filepath, 'w') as destination:
        destination.writelines(datapackage)


def read_in_datafile(path_to_datafile: str, config: Dict) -> Amply:
    """Read in a datafile using the Amply parsing class

    Arguments
    ---------
    path_to_datafile: str
    config: Dict
    """
    parameter_definitions = load_parameter_definitions(config)
    datafile_parser = Amply(parameter_definitions)
    logger.debug(datafile_parser)

    with open(path_to_datafile, 'r') as datafile:
        datafile_parser.load_file(datafile)

    return datafile_parser


def convert_amply_to_dataframe(datafile_parser, config) -> Dict[str, pd.DataFrame]:
    """Converts an amply parser to a dict of pandas dataframes

    Arguments
    ---------
    datafile_parser : Amply
    config : Dict

    Returns
    -------
    dict of pandas.DataFrame
    """

    dict_of_dataframes = {}

    for name in datafile_parser.symbols.keys():
        logger.debug("Extracting data for %s", name)
        if config[name]['type'] == 'param':
            indices = config[name]['indices']
            indices_dtypes = [config[index]['dtype'] for index in indices]
            indices.append('VALUE')
            indices_dtypes.append('float')

            raw_data = datafile_parser[name].data
            data = convert_amply_data_to_list(raw_data)
            df = pd.DataFrame(data=data, columns=indices)
            try:
                dict_of_dataframes[name] = check_datatypes(df, config, name)
            except ValueError as ex:
                msg = "Validation error when checking datatype of {}: {}".format(name, str(ex))
                raise ValueError(msg)
        elif config[name]['type'] == 'set':
            data = datafile_parser[name].data
            logger.debug(data)

            indices = ['VALUE']
            df = pd.DataFrame(data=data, columns=indices, dtype=config[name]['dtype'])
            dict_of_dataframes[name] = check_set_datatype(df, config, name)
        logger.debug("\n%s\n", dict_of_dataframes[name])

    return dict_of_dataframes


def convert_amply_data_to_list(amply_data: Dict) -> List[List]:
    """Flattens a dictionary into a list of lists

    Arguments
    ---------
    amply_data: dict
    """

    data = []

    raw_data = flatten(amply_data)
    for key, value in raw_data.items():
        data.append(list(key) + [value])

    return data


def load_parameter_definitions(config: dict) -> str:
    """Load the set and parameter dimensions into datafile parser

    Returns
    -------
    str
    """
    elements = ""

    for name, attributes in config.items():
        if attributes['type'] == 'param':
            elements += 'param {} {};\n'.format(name, "{" + ",".join(attributes['indices']) + "}")
        elif attributes['type'] == 'set':
            elements += 'set {};\n'.format(name)

    logger.debug(elements)
    return elements


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)

    path_to_datafile = sys.argv[1]
    path_to_datapackage = sys.argv[2]

    convert_file_to_package(path_to_datafile, path_to_datapackage)



import io

import pandas as pd

from otoole.preprocess.excel_to_osemosys import read_config
from otoole.preprocess.narrow_to_datafile import write_parameter, write_set


class TestDataFrameWriting:

    def test_write_empty_parameter_with_defaults(self):

        data = []

        df = pd.DataFrame(data=data, columns=['REGION', 'FUEL', 'VALUE'])

        stream = io.StringIO()
        write_parameter(stream, df, 'test_parameter', 0)

        stream.seek(0)
        expected = ['param default 0 : test_parameter :=\n',
                    ';\n']
        actual = stream.readlines()

        for actual_line, expected_line in zip(actual, expected):
            assert actual_line == expected_line

    def test_write_parameter_as_tabbing_format(self):

        data = [['SIMPLICITY', 'BIOMASS', 0.95969],
                ['SIMPLICITY', 'ETH1', 4.69969]]

        df = pd.DataFrame(data=data, columns=['REGION', 'FUEL', 'VALUE'])

        stream = io.StringIO()
        write_parameter(stream, df, 'test_parameter', 0)

        stream.seek(0)
        expected = ['param default 0 : test_parameter :=\n',
                    'SIMPLICITY BIOMASS 0.95969\n',
                    'SIMPLICITY ETH1 4.69969\n',
                    ';\n']
        actual = stream.readlines()

        for actual_line, expected_line in zip(actual, expected):
            assert actual_line == expected_line

    def test_write_set(self):

        data = [['BIOMASS'],
                ['ETH1']]

        df = pd.DataFrame(data=data, columns=['VALUE'])

        stream = io.StringIO()
        write_set(stream, df, 'TECHNOLOGY')

        stream.seek(0)
        expected = ['set TECHNOLOGY :=\n',
                    'BIOMASS\n',
                    'ETH1\n',
                    ';\n']
        actual = stream.readlines()

        for actual_line, expected_line in zip(actual, expected):
            assert actual_line == expected_line


class TestConfig:

    def test_read_config(self):

        actual = read_config()
        expected = {
                        'default': 0,
                        'dtype': 'float',
                        'indices': ['REGION', 'FUEL', 'YEAR'],
                        'type': 'param',
                    }
        assert actual['AccumulatedAnnualDemand'] == expected



"""Provides a command line interface to the ``otoole`` package

The key functions are **convert**, **cplex** and **viz**.

The ``convert`` command allows convertion of multiple different OSeMOSYS input formats including
from/to csv, an AMPL format datafile, a Tabular Data Package, a folder of CSVs, an Excel workbook with
one tab per parameter, an SQLite database

The ``cplex`` command provides access to scripts which transform and process a CPLEX solution file
into a format which is more readily processed - either to CBC or CSV format.

The ``validate`` command checks the technology and fuel names of a tabular data package against a standard
or user defined configuration file.

The **viz** command allows you to produce a Reference Energy System diagram from a Tabular Data Package.

Example
-------

Ask for help on the command line::

    >>> $ otoole --help
    usage: otoole [-h] [--verbose] [--version] {convert,cplex,validate,viz} ...

    otoole: Python toolkit of OSeMOSYS users

    positional arguments:
    {convert,cplex,validate,viz}
        convert             Convert from one input format to another
        cplex               Process a CPLEX solution file
        validate            Validate an OSeMOSYS model
        viz                 Visualise the model

    optional arguments:
    -h, --help            show this help message and exit
    --verbose, -v         Enable debug mode
    --version, -V         The version of otoole

"""
import argparse
import logging
import sys

from otoole import __version__, read_packaged_file
from otoole.preprocess import convert_file_to_package, create_datafile, create_datapackage, generate_csv_from_excel
from otoole.preprocess.create_datapackage import convert_datapackage_to_sqlite
from otoole.results.convert import convert_cplex_file
from otoole.validate import main as validate
from otoole.visualise import create_res


def validate_model(args):
    file_format = args.format

    if args.config:
        config = read_packaged_file(args.config)
        validate(file_format, args.filepath, config)
    else:
        validate(file_format, args.filepath)


def cplex2cbc(args):
    convert_cplex_file(args.cplex_file, args.output_file, args.start_year,
                       args.end_year, args.output_format)


def conversion_matrix(args):
    """Convert from one format to another

    Implemented conversion functions::

        from\to     ex cs dp sq df
        --------------------------
        excel       -- yy
        csv         nn -- yy nn nn
        datapackage nn ?? -- yy yy
        sql         nn       -- yy
        datafile    nn ?? yy    --

    """

    msg = "Conversion from {} to {} is not yet implemented".format(args.from_format, args.to_format)

    if args.from_format == 'datafile':

        if args.to_format == 'datapackage':
            convert_file_to_package(args.from_path, args.to_path)
        else:
            raise NotImplementedError(msg)

    elif (args.from_format == 'datapackage'):

        if args.to_format == 'sql':
            convert_datapackage_to_sqlite(args.from_path, args.to_path)
        elif args.to_format == 'datafile':
            create_datafile(args.from_path, args.to_path)
        else:
            raise NotImplementedError(msg)

    elif args.from_format == 'sql':

        if args.to_format == 'datafile':
            create_datafile(args.from_path, args.to_path, sql=True)
        else:
            raise NotImplementedError(msg)

    elif args.from_format == 'csv':
        if args.to_format == 'datapackage':
            create_datapackage(args.from_path, args.to_path)
        else:
            raise NotImplementedError(msg)

    elif args.from_format == 'excel':
        if args.to_format == 'csv':
            generate_csv_from_excel(args.from_path, args.to_path)
        else:
            raise NotImplementedError(msg)

    else:
        raise NotImplementedError(msg)


def datapackage2res(args):
    create_res(args.datapackage, args.resfile)


def get_parser():
    parser = argparse.ArgumentParser(description="otoole: Python toolkit of OSeMOSYS users")

    parser.add_argument('--verbose', '-v', help='Enable debug mode', action='count', default=0)
    parser.add_argument('--version', '-V', action='version', version=__version__,
                        help='The version of otoole')

    subparsers = parser.add_subparsers()

    # Parser for conversion
    convert_parser = subparsers.add_parser('convert', help='Convert from one input format to another')
    convert_parser.add_argument('from_format', help='Input data format to convert from',
                                choices=sorted(['datafile', 'datapackage', 'sql', 'excel', 'csv']))
    convert_parser.add_argument('to_format', help='Input data format to convert to',
                                choices=sorted(['datafile', 'datapackage', 'sql', 'csv']))
    convert_parser.add_argument('from_path', help="Path to file or folder to convert from")
    convert_parser.add_argument('to_path', help='Path to file or folder to convert to')
    convert_parser.set_defaults(func=conversion_matrix)

    # Parser for the CPLEX related commands
    cplex_parser = subparsers.add_parser('cplex',
                                         help='Process a CPLEX solution file')

    cplex_parser.add_argument("cplex_file",
                              help="The filepath of the OSeMOSYS cplex output file")
    cplex_parser.add_argument("output_file",
                              help="The filepath of the converted file that will be written")
    cplex_parser.add_argument("-s", "--start_year", type=int, default=2015,
                              help="Output only the results from this year onwards")
    cplex_parser.add_argument("-e", "--end_year", type=int, default=2070,
                              help="Output only the results upto and including this year")
    cplex_parser.add_argument('output_format', choices=['csv', 'cbc'], default='cbc')
    cplex_parser.set_defaults(func=cplex2cbc)

    # Parser for validation
    valid_parser = subparsers.add_parser('validate', help='Validate an OSeMOSYS model')
    valid_parser.add_argument('format', help='The format of the OSeMOSYS model to validate')
    valid_parser.add_argument('filepath', help='Path to the OSeMOSYS model to validate')
    valid_parser.add_argument('--config', help='Path to a user-defined validation-config file')
    valid_parser.set_defaults(func=validate_model)

    # Parser for visualisation
    viz_parser = subparsers.add_parser('viz', help='Visualise the model')
    viz_subparsers = viz_parser.add_subparsers()

    res_parser = viz_subparsers.add_parser('res', help='Generate a reference energy system')
    res_parser.add_argument('datapackage', help='Path to model datapackage')
    res_parser.add_argument('resfile', help='Path to reference energy system')
    res_parser.set_defaults(func=datapackage2res)

    return parser


def main():

    parser = get_parser()
    args = parser.parse_args(sys.argv[1:])

    if args.verbose >= 1 and args.verbose < 2:
        logging.basicConfig(level=logging.INFO)
    if args.verbose > 2:
        logging.basicConfig(level=logging.DEBUG)

    def exception_handler(exception_type, exception, traceback, debug_hook=sys.excepthook):
        if args.verbose:
            debug_hook(exception_type, exception, traceback)
        else:
            print("{}: {}".format(exception_type.__name__, exception.message))

    sys.excepthook = exception_handler

    if 'func' in args:
        args.func(args)
    else:
        parser.print_help()



from .excel_to_osemosys import generate_csv_from_excel
from .create_datapackage import main as create_datapackage
from .narrow_to_datafile import main as create_datafile
from .datafile_to_datapackage import convert_file_to_package

__all__ = ['generate_csv_from_excel', 'create_datapackage', 'create_datafile', 'convert_file_to_package',
           'create_datapackage_from_datafile']



import pandas as pd
from pulp.amply import Amply

from otoole.preprocess.datafile_to_datapackage import (
    convert_amply_data_to_list,
    convert_amply_to_dataframe,
    load_parameter_definitions
)


def test_amply():

    Amply("""set REGION;
          # set REGION := SIMPLICITY;
          set TECHNOLOGY;
          set TECHNOLOGY := ETHPLANT GAS_EXTRACTION;
          set MODE_OF_OPERATION;
          set MODE_OF_OPERATION := 1 2;
          set YEAR;
          set YEAR := 2014;
          end;""")


def test_convert_amply_to_dataframe():

    config = {'VariableCost': {'type': 'param',
                               'indices': ['REGION', 'TECHNOLOGY', 'MODE_OF_OPERATION', 'YEAR'],
                               'dtype': 'float',
                               'default': 0},
              'REGION': {'type': 'set', 'dtype': 'str'},
              'YEAR': {'dtype': 'int', 'type': 'set'},
              'MODE_OF_OPERATION': {'dtype': 'int', 'type': 'set'},
              'TECHNOLOGY': {'dtype': 'str', 'type': 'set'}}

    amply = Amply("""set REGION;
                     set REGION := SIMPLICITY;
                     set TECHNOLOGY;
                     set TECHNOLOGY := ETHPLANT GAS_EXTRACTION;
                     set MODE_OF_OPERATION;
                     set MODE_OF_OPERATION := 1 2;
                     set YEAR;
                     set YEAR := 2014;""")
    amply.load_string(
        "param VariableCost {REGION,TECHNOLOGY,MODE_OF_OPERATION,YEAR};")
#     amply.load_string("""param default 0 : VariableCost :=
# SIMPLICITY ETHPLANT 1 2014 2.89
# SIMPLICITY ETHPLANT 2 2014 999999.0
# SIMPLICITY GAS_EXTRACTION 1 2014 7.5
# SIMPLICITY GAS_EXTRACTION 2 2014 999999.0""")
    amply.load_string("""
param VariableCost default 0.0001 :=
[SIMPLICITY,ETHPLANT,*,*]:
2014 :=
1 2.89
2 999999.0
[SIMPLICITY,GAS_EXTRACTION,*,*]:
2014 :=
1 7.5
2 999999.0;""")
    actual = convert_amply_to_dataframe(amply, config)
    expected = pd.DataFrame(
        data=[['SIMPLICITY', 'ETHPLANT', 1, 2014, 2.89],
              ['SIMPLICITY', 'ETHPLANT', 2, 2014, 999999.0],
              ['SIMPLICITY', 'GAS_EXTRACTION', 1, 2014, 7.5],
              ['SIMPLICITY', 'GAS_EXTRACTION', 2, 2014, 999999.0]],
        columns=['REGION', 'TECHNOLOGY', 'MODE_OF_OPERATION', 'YEAR', 'VALUE'])

    pd.testing.assert_frame_equal(actual['VariableCost'], expected)


def test_convert_amply_data_to_list_of_lists():

    data = {'SIMPLICITY': {
        'ETHPLANT': {
            1.0: {2014.0: 2.89},
            2.0: {2014.0: 999999.0}},
        'GAS_EXTRACTION': {
            1.0: {2014.0: 7.5},
            2.0: {2014.0: 999999.0}},
                            }
            }
    expected = [['SIMPLICITY', 'ETHPLANT', 1.0, 2014.0, 2.89],
                ['SIMPLICITY', 'ETHPLANT', 2.0, 2014.0, 999999.0],
                ['SIMPLICITY', 'GAS_EXTRACTION', 1.0, 2014.0, 7.5],
                ['SIMPLICITY', 'GAS_EXTRACTION', 2.0, 2014.0, 999999.0]]
    actual = convert_amply_data_to_list(data)
    assert actual == expected


def test_load_parameters():

    config = {
        'TestParameter': {'type': 'param',
                          'indices': ['index1', 'index2']}}

    actual = load_parameter_definitions(config)
    expected = 'param TestParameter {index1,index2};\n'
    assert actual == expected


def test_load_sets():

    config = {
        'TestSet': {'type': 'set'}}

    actual = load_parameter_definitions(config)
    expected = 'set TestSet;\n'
    assert actual == expected



"""Visualise the reference energy system
"""
import logging
import os
import sys
from typing import Dict, List, Tuple

import networkx as nx  # mypy: ignore
from datapackage import Package

logger = logging.getLogger(__name__)


def load_datapackage(path_to_datapackage: str) -> Package:

    package = Package(path_to_datapackage)

    return package


def extract_nodes(package_rows: List[List], node_type='technology',
                  color='red', shape='circle') -> List[Tuple[str, Dict]]:
    """Add nodes from a Tabular Data Table

    Arguments
    ---------
    package_rows : list of dict
    node_type : str, default='technology'
    color : str, default='red'
    shape : str, default='circle'

    Returns
    -------
    list of tuple
        A list of nodes with attributes
    """
    nodes = [(x[0], {'type': node_type,
                     'fillcolor': color, 'shape': shape,
                     'style': 'filled',
                     'label': x[0]}
              )
             for x in package_rows]

    return nodes


def add_fuel(package_rows: List[List]) -> List[Tuple[str, Dict]]:
    """Add fuel nodes

    Arguments
    ---------
    package_rows : list of dict

    Returns
    -------
    list of tuple
        A list of node names along with a dict of node attributes
    """
    nodes = [(x[0],
              {'type': 'fuel', 'style': 'filled', 'shape': 'circle', 'height': 0.1, 'width': 0.1, 'label': ""}
              )
             for x in package_rows]
    return nodes


def extract_edges(package_rows: List[Dict], from_column: str,
                  to_column: str, parameter_name: str,
                  directed: bool = True) -> List[Tuple[str, str, Dict]]:
    """Add edges from a Tabular Data Table

    Arguments
    ---------
    package_rows : list of dict
    from_column : str
        The name of the column to use as source of the edge
    to_column: str
        The name of the column to use as a destination of the edge
    parameter_name: str
        The name of the parameter
    directed: bool, default=True
        Specifies whether the edge should have an arrow or not

    Returns
    -------
    list of tuple
        A list of edges with from/to nodes names and edge attributes
    """
    if directed:
        edges = [(x[from_column], x[to_column],
                 {parameter_name: float(x['VALUE']), 'dir': 'none'})
                 for x in package_rows]
    else:
        edges = [(x[from_column], x[to_column],
                 {parameter_name: float(x['VALUE']), 'xlabel': x[from_column]})
                 for x in package_rows]

    return edges


def create_graph(datapackage: Package):
    """Creates a graph of technologies and fuels

    Arguments
    ---------
    datapackage : datapackage.Package

    Returns
    -------
    graph
        networkx.DiGraph
    """

    technologies = datapackage.get_resource('TECHNOLOGY').read()
    storage = datapackage.get_resource('STORAGE').read()
    fuel = datapackage.get_resource('FUEL').read()
    emission = datapackage.get_resource('EMISSION').read()

    nodes = extract_nodes(technologies, shape='rectangle', color='yellow')
    nodes += extract_nodes(storage, node_type='storage', shape='triangle', color='lightblue')
    nodes += add_fuel(fuel)
    nodes += extract_nodes(emission, node_type='emission', color='grey')
    nodes += [('AnnualDemand',
               {'type': 'demand', 'fillcolor': 'green',
                'label': 'AccumulatedAnnualDemand',
                'style': 'filled'}
               )
              ]

    input_activity = datapackage.get_resource('InputActivityRatio').read(keyed=True)
    output_activity = datapackage.get_resource('OutputActivityRatio').read(keyed=True)
    emission_activity = datapackage.get_resource('EmissionActivityRatio').read(keyed=True)
    tech2storage = datapackage.get_resource('TechnologyToStorage').read(keyed=True)
    techfromstorage = datapackage.get_resource('TechnologyFromStorage').read(keyed=True)
    acc_demand = datapackage.get_resource('AccumulatedAnnualDemand').read(keyed=True)
    spec_demand = datapackage.get_resource('SpecifiedAnnualDemand').read(keyed=True)

    edges = extract_edges(input_activity, 'FUEL', 'TECHNOLOGY', 'input_ratio', directed=False)

    edges += extract_edges(output_activity, 'TECHNOLOGY', 'FUEL', 'output_ratio')

    edges += extract_edges(emission_activity, 'TECHNOLOGY', 'EMISSION', 'emission_ratio')

    edges += extract_edges(tech2storage, 'TECHNOLOGY', 'STORAGE', 'input_ratio')

    edges += extract_edges(techfromstorage, 'STORAGE', 'TECHNOLOGY', 'ouput_ratio')

    edges += [(x['FUEL'], 'AnnualDemand',
              {'Demand': float(x['VALUE']), 'xlabel': x['FUEL']})
              for x in acc_demand]
    edges += [(x['FUEL'], 'AnnualDemand',
              {'Demand': float(x['VALUE']), 'xlabel': x['FUEL']})
              for x in spec_demand]

    graph = build_graph(nodes, edges)

    return graph


def create_res(path_to_datapackage: str, path_to_resfile: str):
    """Create a reference energy system diagram from a Tabular Data Package

    Arguments
    ---------
    path_to_datapackage : str
        The path to the ``datapackage.json``
    path_to_resfile : str
        The path to the image file to be created
    """
    logger.debug(path_to_resfile, path_to_resfile)
    package = load_datapackage(path_to_datapackage)

    graph = create_graph(package)
    draw_graph(graph, path_to_resfile)


def draw_graph(graph, path_to_resfile):
    """Layout the graph and write it to disk

    Uses pygraphviz to set some graph attributes, layout the graph
    and write it to disk

    Arguments
    ---------
    path_to_resfile : str
        The file path of the PNG image file that will be created
    """
    for node, attributes in graph.nodes.data():
        logger.debug("%s: %s", node, attributes)
    for source, sink, attributes in graph.edges.data():
        logger.debug("%s-%s: %s", source, sink, attributes)

    filename, ext = os.path.splitext(path_to_resfile)
    nx.write_graphml(graph, filename + '.graphml')
    dot_graph = nx.nx_pydot.to_pydot(graph)

    dot_graph.set('rankdir', 'LR')
    dot_graph.set('splines', 'ortho')
    dot_graph.set('concentrate', 'true')

    image_format = ext.strip(".")

    image = dot_graph.create(prog='dot', format=image_format)
    with open(path_to_resfile, 'wb') as image_file:
        image_file.write(image)


def build_graph(nodes: List[Tuple[str, Dict]], edges: List[Tuple[str, str, Dict]]) -> nx.DiGraph:
    """Builds the graph using networkx

    Arguments
    ---------
    nodes : list
        A list of node tuples
    edges : list
        A list of edge tuples

    Returns
    -------
    networkx.DiGraph
        A directed graph representing the reference energy system
    """
    graph = nx.DiGraph()
    graph.add_nodes_from(nodes)
    graph.add_edges_from(edges)
    return graph


if __name__ == '__main__':

    logging.basicConfig(level=logging.DEBUG)
    path_to_datapackage = sys.argv[1]
    path_to_resfile = sys.argv[2]
    create_res(path_to_datapackage, path_to_resfile)
